package SPVM::Hash::Entry;

1;

=head1 Name

SPVM::Hash::Entry - Hash entry

=head1 Usage
  
  use Hash::Entry;

=head1 Description

L<Hash::Entry|SPVM::Hash::Entry> is used for L<Hash|SPVM::Hash> internally.
